from argparse import ArgumentParser
import sys
sys.path.append('../../')

from transformers.trainer_utils import set_seed
set_seed(42)


import numpy as np
import matplotlib.pyplot as plt
import torch
import pyro.distributions as dist


from src.utils import read_tensor_data, get_array, plot_curves, show, get_config

def main(IG_a=1e-1, IG_b=1e-3, minnesota_lambda=0.5, minnesota_gamma=0.9,
         nn_prior_var=0.05, nb_epochs=500):
    device = 'cpu'
    log_yield = read_tensor_data('../../data/processed/curve_data.txt')


    horizons = [5, 20, 60, 120]
    train_start, train_end, test_end = [0, 2833, 3589]
    maturities = [21, 42, 63, 126, 252, 504, 756, 1008, 1260, 1512, 1764]

    print(maturities)
    print(horizons)

    tr_log_yield = log_yield[train_start:train_end]
    print(tr_log_yield.shape)

    te_log_yield = log_yield[train_end:test_end]
    te_log_yield.shape

    tr_y = torch.tensor(tr_log_yield[:, maturities], dtype=torch.float64).to(device)
    te_y = torch.tensor(te_log_yield[:, maturities], dtype=torch.float64).to(device)

    tr_y_norm = tr_y - tr_y.mean(axis=0)

    def normalize(x):     
        y_tr_bar = tr_y.mean(axis=0)
        if torch.is_tensor(x):
            return x - y_tr_bar
        return x - get_array(y_tr_bar)

    def denormalize(x):       
        y_tr_bar = tr_y.mean(axis=0)
        if torch.is_tensor(x):
            return x + y_tr_bar
        return x + get_array(y_tr_bar)


    m = torch.tensor(maturities, dtype=torch.float64)[..., None]/252
    m = m.to(device)

    class BasisLearner(torch.nn.Module):
        '''
            Aprende uma base de k funções f:[1, M] -> [0, 1]
        '''

        def __init__(self, intercept_term=True):
            super().__init__()
                        
            self.mlp = torch.nn.Sequential(
                torch.nn.Linear(1, 300),
                torch.nn.Tanh(),
                torch.nn.Linear(300, 300),
                torch.nn.Tanh(),
                torch.nn.Linear(300, 3 if intercept_term else 4),
                torch.nn.Sigmoid()
            )
            
            self.intercept_term = intercept_term
        
        def forward(self, m):
            basis_fns = self.mlp(m)
            intercept_term = torch.ones_like(basis_fns[:, 0:1])

            if self.intercept_term:
                return torch.cat([intercept_term, basis_fns], axis=1)
            return basis_fns

    def get_pretrained_basis_learner(intercept=True, tol=1e-6):
        L = 1.6
        L2 = 1.0
        nss_obs_matrix = []

        for maturity in (maturities):
            maturity = maturity/252
            nss_obs_matrix.append(
                [1, (1-np.exp(-(maturity*L)))/(maturity*L), (1-np.exp(-(maturity*L)))/(maturity*L) - np.exp(-(maturity*L)), 
                (1-np.exp(-(maturity*L2)))/(maturity*L2) - np.exp(-(maturity*L2))]
            )
        nss_obs_matrix = torch.tensor(nss_obs_matrix, dtype=torch.float64).T
        nss_obs_matrix = torch.tensor(nss_obs_matrix, requires_grad=False)
        
        basis_learner = BasisLearner(intercept_term=intercept).double().to(device)
        
        pre_opt = torch.optim.Adam(basis_learner.parameters(), lr=2e-4)
        criterion = torch.nn.MSELoss()
        
        for i in range(10_000):
            pre_opt.zero_grad()

            output = basis_learner(m)
            pre_loss = criterion(output[:, 0:4], nss_obs_matrix.T[:, 0:4].to(device))

            pre_loss.backward()
            pre_opt.step()

            #if (i+1) % 100 == 0:
            #    print('%d/2000: %0.6f' % (i+1, pre_loss.item()))
            
            if pre_loss.item() < tol:
                break
        return basis_learner


    def plot_basis_fns(basis_learner, fpath):
        vals = get_array(basis_learner(torch.arange(1, 2500, dtype=torch.float64, device=device)[...,None]/252))

        with torch.no_grad():
            plt.plot(torch.arange(1, 2500, dtype=torch.float64)/252, vals)
        
        plt.savefig(f'imgs/{fpath}.png')
        np.savetxt(f'arrays/{fpath}.txt', vals)

    # Cria basis learner
    basis_learner = BasisLearner(intercept_term=True).double().to(device)
    basis_learner.load_state_dict(torch.load(open(f'./tmp/blank_basis_learner_4.pk', 'rb')))

    def ig_pdf(x):
        return IG_b**IG_a * x**(-IG_a-1) * np.exp(-IG_b/x)

    def minnesota_prior(trans_sigma, trans_matrix):
        # IG prior
        log_pdf = -(IG_a + 1)*torch.log(trans_sigma) - IG_b/trans_sigma
        log_pdf = log_pdf.sum()
        
        # V matrix
        trans_sigma_col = trans_sigma[..., None]
        V = trans_sigma_col.div(trans_sigma_col.T)*(minnesota_lambda**2)
        V = V * ((1 - torch.eye(*trans_sigma.shape, device=device))*(minnesota_gamma**2) + torch.eye(*trans_sigma.shape, device=device))
        
        # Normal prior pdf
        demean_trans_matrix = (trans_matrix - torch.eye(*trans_matrix.shape, device=device)).view(-1, 1)
        
        log_pdf += -1/2 * torch.log(V).sum() - 1/2 * torch.sum((demean_trans_matrix/(V.view(-1)))**2)
        
        return log_pdf

    def weight_L2(basis_learner):
        return -1/2 * 1/nn_prior_var * sum(
            (t**2).sum() for name, t in basis_learner.named_parameters()
            if 'bias' not in name
        )

    from torch.optim import Adam
    from pyro import distributions as dist

    init_dist_mu = torch.tensor([0.], requires_grad=True, dtype=torch.float64, device=device)
    init_dist_sigma = torch.tensor([1.], requires_grad=True, dtype=torch.float64, device=device)
    init_dist = dist.Normal(init_dist_mu, init_dist_sigma).expand([4]).to_event()

    # F
    trans_matrix = torch.tensor(np.eye(4), requires_grad=True, dtype=torch.float64, device=device)

    # E(w_t)
    mu_trans = torch.tensor(np.random.randn(4)*0, requires_grad=False, dtype=torch.float64, device=device)

    # P
    trans_sigma = torch.tensor([1e-3]*(4), requires_grad=True, dtype=torch.float64, device=device)

    # w_t
    trans_dist = dist.Normal(mu_trans, trans_sigma).expand([4]).to_event()

    # observation_matrix = nss_obs_matrix
    observation_matrix = basis_learner(m).T
    observation_sigma = torch.tensor([5e-4]*len(maturities), dtype=torch.float64, device=device)
    observation_dist = dist.Normal(0., observation_sigma).expand([len(maturities)]).to_event(1)

    g = dist.GaussianHMM(initial_dist=init_dist, transition_matrix=trans_matrix, transition_dist=trans_dist, observation_matrix=observation_matrix, 
                        observation_dist=observation_dist, duration=train_end-train_start)


    opt = Adam([
        {'params': list(basis_learner.parameters()), 'lr': 1e-3},
        {'params': [trans_sigma, init_dist_sigma], 'lr': 1e-4},
        {'params': [observation_sigma], 'lr': 5e-5},
        {'params': [trans_matrix, mu_trans, init_dist_mu], 'lr': 1e-3},
    ])


    for i in range(nb_epochs):
        # Zera gradiente
        opt.zero_grad()
        
        # Obtém nova base de funções
        observation_matrix = basis_learner(m).T
        
        # Define modelo de espaço de estados
        g = dist.GaussianHMM(initial_dist=init_dist, transition_matrix=trans_matrix, transition_dist=trans_dist,
                            observation_matrix=observation_matrix, observation_dist=observation_dist, duration=train_end-train_start)
        
        # Obtém -p(theta|x, y)
        loss = -g.log_prob(tr_y_norm) - minnesota_prior(trans_sigma, trans_matrix) - weight_L2(basis_learner)
        loss.backward(retain_graph=True)
        
        # Executa um passo do otimizador
        opt.step()
        
        # Logging
        if (i+1) % 5 == 0:
            print(f'{i}/{nb_epochs} : {loss.item():,}')
            print(np.round(get_array(trans_sigma), 4))
            print(np.round(get_array(observation_sigma), 4))
            print('-'*10)
        
        with torch.no_grad():
            observation_sigma += (observation_sigma < 0)*(-1*observation_sigma.min() + 1e-7)
            trans_sigma += (trans_sigma < 0)*(-1*trans_sigma.min() + 1e-7)
        
        #trans_matrix.requires_grad = True

    print(f'mu_trans = {np.round(get_array(mu_trans), 4)}')

    print(f'trans_sigma = {np.round(get_array(trans_sigma), 4)}')

    print(f'observation_sigma = {np.round(get_array(observation_sigma), 4)}')

    print(f'trans_matrix = {np.round(trans_matrix.detach().numpy(), 3)}')

    

    # with torch.no_grad():
    #    np.savetxt('../Basis Learner Model/Priors/plot_data/NN4_basis_y_vals_from_21.txt', 
    #               get_array(basis_learner(torch.arange(1, 2500, dtype=torch.float64, device=device)[..., None]/252))
    #    )

    from tqdm import tqdm
    from pykalman import KalmanFilter


    kf = KalmanFilter(transition_matrices=trans_matrix.detach().numpy(),
                    observation_matrices=observation_matrix.detach().numpy().T,
                    transition_covariance=np.diag(trans_sigma.detach().numpy()),
                    observation_covariance=np.diag(observation_sigma.detach().numpy()),
                    initial_state_mean=get_array(g.filter(tr_y_norm[0:1]).mean),
                    initial_state_covariance=np.diag(get_array(g.filter(tr_y_norm[0:1]).variance)))

    filter_states, filter_covs = kf.filter(get_array(tr_y_norm))
    smoother_states, smoother_covs = kf.smooth(get_array(tr_y_norm))

    #np.savetxt('../Basis Learner Model/Priors/plot_data/factors_nnss4_final_from_21.txt', smoother_states)


    def get_curve_from_factors(basis_learner, factors):
        assert len(factors.shape) > 1
        
        with torch.no_grad():
            observation_matrix = basis_learner(m).T
        try:
            return denormalize(factors@get_array(observation_matrix))
        except:
            import pdb;
            pdb.set_trace()


    def plot_i(i):
        plt.figure(figsize=(15, 5))
        plt.plot(maturities, denormalize(filter_states[i]@get_array(observation_matrix)), 'o--')
        plt.plot(maturities, get_array(tr_y[i]))
        mi, ma = get_array(tr_y)[i].min(), get_array(tr_y)[i].max()
        plt.ylim(mi - (ma-mi)*0.1, ma + (ma-mi)*0.1)


    def forecast(factors, transition_matrix, h=1):
        trans_matrix_np = get_array(trans_matrix)
        last_pred = factors[-1:]
        predictions = [None]*h
        
        for i in range(h):
            last_pred = last_pred@trans_matrix_np
            predictions[i] = last_pred
        500
        return np.concatenate(predictions)


    horizon = 60

    get_curve_from_factors(basis_learner, filter_states[0:2])

    predicted_factors = forecast(filter_states, trans_matrix, h=horizon)

    predicted_curves = get_curve_from_factors(basis_learner, predicted_factors)


    def recursive_forecast(train_tensor, test_tensor, horizons, window_step=1):
        
        forecast_results = {h: {} for h in horizons}
        
        for h in tqdm(horizons):
            current_state, current_cov = smoother_states[-1], smoother_covs[-1]
            
            for i in tqdm(np.arange(0, test_tensor.shape[0]-h, window_step)):
                current_index = i
                forecast_results[h][i] = {}
                
                current_dataset = torch.cat([train_tensor, normalize(test_tensor)[0:current_index]])
                
                # Registra index e valor real
                forecast_results[h][i]['current_index'] = current_index
                forecast_results[h][i]['real'] = get_array(test_tensor[current_index+h-1])
                
                # Random walk
                forecast_results[h][i]['rw'] = denormalize(get_array(current_dataset))[-1]
                
                
                with torch.no_grad():
                    # Filtro de Kalman
                    current_state, current_cov = kf.filter_update(current_state, current_cov,
                                                                get_array(current_dataset[-1]))
                    current_state, current_cov = np.array(current_state), np.array(current_cov)
                    
                    # Preve novos estados
                    new_state = forecast(current_state.reshape(1, -1), trans_matrix, h)
                    500
                    # Obtém a curva do último estado
                    forecasted = get_curve_from_factors(basis_learner, new_state[-1:]).squeeze()
                
                forecast_results[h][i]['factor_forecast'] = new_state
                forecast_results[h][i]['forecast'] = forecasted
        return forecast_results


    def get_metrics(forecast_results, horizons):
        forecast_metrics = {h: {} for h in horizons}
        
        for horizon in horizons:
            n = len(forecast_results[horizon])
            rw_mse, mse, mae = 0., 0., 0.
            
            for dict_step in forecast_results[horizon].values():
                mse += (dict_step['real'] - dict_step['forecast'])**2
                mae += np.abs(dict_step['real'] - dict_step['forecast'])
                rw_mse += (dict_step['real'] - dict_step['rw'])**2

            forecast_metrics[horizon]['mae'] = (mae/n)
            forecast_metrics[horizon]['mse'] = (mse/n)
            forecast_metrics[horizon]['rw_mse'] = (rw_mse/n)
            forecast_metrics[horizon]['rw_rmse'] = np.sqrt(rw_mse/n)
            forecast_metrics[horizon]['rmse'] = (np.sqrt(mse/n))
        return forecast_metrics

    forecast_results = recursive_forecast(tr_y_norm, te_y, [5, 20, 60, 20*6])
    forecast_metrics = get_metrics(forecast_results, [5, 20, 60, 120])

    def export_table(mse_list):
        lista = np.round(np.sqrt(mse_list)*1e3, 2)
        print(' & '.join([str(x) for x in lista]))

    def export_avg(mse_list):
        lista = np.round(np.sqrt(mse_list)*1e3, 2)
        return np.mean(lista)

    # Modelo atual vs Competidor
    import pandas as pd
    from pathlib import Path

    df = []
    model_name = f'NNSS_a_{IG_a:e}_b_{IG_b:e}_la_{minnesota_lambda:e}_ga_{minnesota_gamma:e}_va_{nn_prior_var}_ep_{nb_epochs}'
    
    plot_basis_fns(basis_learner, model_name)

    for h in [5, 20, 60, 120]:
        df.append(
            [model_name, h, *(np.sqrt(forecast_metrics[h]['mse'])*1e3).tolist(),
            np.mean(np.sqrt(forecast_metrics[h]['mse'])*1e3), np.std(np.sqrt(forecast_metrics[h]['mse'])*1e3), ]
        )

    df = pd.DataFrame(df, columns=['model_name', 'horizon', *[str(int(m)) for m in np.array(maturities)/21], 'Avg.RMSE', 'Std.RMSE'])

    df['IG_a'] = IG_a
    df['IG_b'] = IG_b
    df['minnesota_lambda'] = minnesota_lambda
    df['minnesota_gamma'] = minnesota_gamma
    df['nn_prior_var'] = nn_prior_var
    df['nb_epochs'] = nb_epochs

    if not Path('./forecasts.csv').exists():
        df.to_csv('./forecasts.csv', index=False)
    else:
        df_antigo = pd.read_csv('./forecasts.csv')
        df = pd.concat([df_antigo, df])
        df.to_csv('./forecasts.csv', index=False)

    print(df)

parser = ArgumentParser()
parser.add_argument('--IG_a', type=float, default=0.1)
parser.add_argument('--IG_b', type=float, default=0.001)
parser.add_argument('--minnesota_lambda', type=float, default=0.5)
parser.add_argument('--minnesota_gamma', type=float, default=0.9)
parser.add_argument('--nn_prior_var', type=float, default=0.05)
parser.add_argument('--nb_epochs', type=int, default=500)

if __name__ == '__main__':
    args = parser.parse_args()
    main(args.IG_a, args.IG_b, args.minnesota_lambda,
          args.minnesota_gamma, args.nn_prior_var, args.nb_epochs)
