from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
import torch

class BasisLearner(torch.nn.Module):
    '''
        Aprende uma base de k funções f:[1, M] -> [0, 1]
    '''

    def __init__(self, nb_factors=4, intercept_term=True, nb_hidden=300):
        super().__init__()
        
        self.nb_factors = nb_factors
        self.nb_hidden = nb_hidden
        
        self.mlp = torch.nn.Sequential(
            torch.nn.Linear(1, nb_hidden),
            torch.nn.Tanh(),
            torch.nn.Linear(nb_hidden, nb_hidden),
            torch.nn.Tanh(),
            torch.nn.Linear(nb_hidden, nb_factors-1 if intercept_term else nb_factors),
            torch.nn.Sigmoid()
        )
        
        self.intercept_term = intercept_term
    
    def forward(self, m):
        basis_fns = self.mlp(m)
        intercept_term = torch.ones_like(basis_fns[:, 0:1])

        if self.intercept_term:
            return torch.cat([intercept_term, basis_fns], axis=1)
        return basis_fns
    
    def load_weights(self, model_path):
        model_path = Path(model_path)
        assert model_path.exists()

        self.load_state_dict(torch.load(open(model_path, 'rb')))

def plot_basis_fns(basis_learner, device='cpu', title=''):
    with torch.no_grad():
        loadings = basis_learner(torch.arange(1, 2500, dtype=torch.float64, device=device)[...,None]/252)
        loadings = loadings.detach().cpu().numpy()
        plt.plot(
            torch.arange(1, 2500, dtype=torch.float64)/252,
            loadings 
        )
        plt.title(title)
        plt.show()
