import json
import numpy as np

def get_metrics(forecast_results, horizons):
    forecast_metrics = {h: {} for h in horizons}
    
    for horizon in horizons:
        n = len(forecast_results[horizon])
        rw_mse, mse, mae = 0., 0., 0.
        
        for dict_step in forecast_results[horizon].values():
            mse += (dict_step['real'] - dict_step['forecast'])**2
            mae += np.abs(dict_step['real'] - dict_step['forecast'])
            rw_mse += (dict_step['real'] - dict_step['rw'])**2

        forecast_metrics[horizon]['mae'] = (mae/n)
        forecast_metrics[horizon]['mse'] = (mse/n)
        forecast_metrics[horizon]['rw_mse'] = (rw_mse/n)
        forecast_metrics[horizon]['rw_rmse'] = np.sqrt(rw_mse/n)
        forecast_metrics[horizon]['rmse'] = (np.sqrt(mse/n))
    return forecast_metrics

def export_table(mse_list):
    lista = np.round(np.sqrt(mse_list)*1e3, 2)
    print(' & '.join([str(x) for x in lista]))

def export_avg(mse_list):
    lista = np.round(np.sqrt(mse_list)*1e3, 2)
    return np.mean(lista)

def save_forecast_results(forecast_metrics, output_file):
    for horizon, metrics in forecast_metrics.items():
        forecast_metrics[horizon] = {
            metric_name: results.tolist() 
            for metric_name, results in metrics.items()
        }
    json.dump(forecast_metrics, open(output_file, 'w', encoding='utf-8'), indent=2)
