import torch
import numpy as np

class InverseGamma():
    def __init__(self, a, b):
        self.a = a
        self.b = b
    
    def log_pdf(self, x):
        log_pdf = -(self.a + 1)*torch.log(x) - self.b/x
        return log_pdf.sum()


class IGMinnesota():
    def __init__(self, IG_a, IG_b, minnesota_lambda, minnesota_gamma, device='cpu'):
        self.IG = InverseGamma(IG_a, IG_b)
        self.minnesota_lambda = minnesota_lambda
        self.minnesota_gamma = minnesota_gamma
        self.device = device

    def log_pdf(self, trans_sigma, trans_matrix):
        log_pdf = self.IG.log_pdf(trans_sigma)

        # Variance matrix
        trans_sigma_col = trans_sigma[..., None]
        V = trans_sigma_col.div(trans_sigma_col.T)*(self.minnesota_lambda**2)
        V = V * ((1 - torch.eye(*trans_sigma.shape, device=self.device))*(self.minnesota_gamma**2) +\
            torch.eye(*trans_sigma.shape, device=self.device))
        
        # Normal prior pdf
        demean_trans_matrix = (trans_matrix - torch.eye(*trans_matrix.shape, device=self.device)).view(-1, 1)
        log_pdf += -1/2 * torch.log(V).sum() - 1/2 * torch.sum((demean_trans_matrix/(V.view(-1)))**2)
        
        return log_pdf


class UncorrelatedNormalNNPrior():
    def __init__(self, nn_prior_variance):
        self.nn_prior_variance = nn_prior_variance
    
    def log_pdf(self, neural_network):
        return -1/2 * 1/self.nn_prior_variance * sum(
            (t**2).sum() for name, t in neural_network.named_parameters()
            if 'bias' not in name
        )
