import pathlib
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from bizdays import load_holidays, Calendar
from typing import Union, List, Tuple, Dict
import torch

anbima_path = str(pathlib.Path(__file__).parent.parent.absolute() / 'data/utils/anbima.txt')
du_cal = Calendar(load_holidays(anbima_path), weekdays=['Sunday', 'Saturday'], name = 'Anbima')


def read_tensor_data(path: str, continuous_compounding: bool = True,
                     torch: bool = False, device: str = 'cpu'):
    pre_arr = np.loadtxt(path)/100
    
    if continuous_compounding:
        pre_arr = np.log(1 + pre_arr)
    
    if torch:
        return torch.tensor(pre_arr, device=device)
    return pre_arr


def get_array(y: torch.Tensor, cuda: bool = True):
    if y.device.type == 'cuda':
        return y.cpu().detach().numpy()
    else:
        return y.detach().numpy()


def plot_curves(y: np.ndarray, 
                cuda: bool = True,
                alpha: float = 1.,
                xlim: Union[None, float] = None):
    if hasattr(y, 'device'):
        curves = get_array(y, cuda)
    else:
        curves = y
    
    if xlim is None:
        xlim = (0, curves.shape[-1])
    
    if len(curves.shape) > 1:
        curve_nb = curves.shape[0]
        colors = plt.cm.viridis(np.linspace(0, 1, curve_nb))
        
        for i in range(curve_nb):
            plt.plot(curves[i], color = colors[i], alpha = alpha)
    else:
        plt.plot(curves)
    plt.xlim(xlim)


def show(*matrix):
    fig, ax = plt.subplots(nrows=1, ncols=len(matrix), figsize=(20,10))
    ax = ax if len(matrix) > 1 else [ax]
        
    cov_min, cov_max = min([m.min() for m in matrix]), max([m.max() for m in matrix])
    
    for i in range(len(matrix)):
        im = ax[i].imshow(matrix[i], vmin=cov_min, vmax=cov_max)
    
    fig.colorbar(im, ax=(ax.ravel().tolist() if len(matrix) > 1 else ax))


def normalize(matrix):
    dinv = np.diag(1/np.sqrt(np.diag(matrix)))
    return dinv@matrix@dinv

